### This script uses existing LPR statistics to segment a corpus into an
### initial vocabulary, see paper for details of the algorithm.



import gc
import copy
import cPickle
from multiprocessing import Process,Queue
from pos_helper import matches_gap_en,matches_gap_fr
from multi_helper import *
from corpus_reader import read_sentence_from_corpus
from LPR_helper import *
gc.disable()



def replace_words_with_id(words,rev_id_dict):
    for i in range(len(words)):
        if words[i] in rev_id_dict:
            words[i] = rev_id_dict[words[i]]
        else:
            words[i] = -2
    return words


def get_ngram_spans(words,ngram_dict):
    spans = [[]]
    i = len(words) - 1
    while i >= 0:
        ID = 0
        stop = False
        if words[i] == -2:
            stop = True
        ID = words[i]
        j = i - 1
        while j >= 0 and not stop:
            if words[j] == -2:
                stop = True
                continue
            else:
                ID = ID << 18 | words[j]
                if not ID in ngram_dict:
                    stop = True
                    continue
            j-= 1
        if i - j > 1:
            if len(spans[-1]) !=0 and j + 1 == spans[-1][-1][0]:
                pass
            else:
                spans[-1].append((j+1,i+1))
        else:
            if spans[-1]:
                spans[-1].sort()
                spans.append([])
        i -= 1

    if not spans[-1]:
        del spans[-1]
    else:
        spans[-1].sort()
    spans.sort()
    return spans

def get_skip_spans(words,org_pos,skip_dict):
    spans = []
    for i in range(0,len(words) - 2):
        j = i + 1
        while j < len(words) and matches_gap(org_pos,i,j):
            m = i - 1
            n = j + 1
            last_m = -1
            stop = False
            while not stop and n <= len(words):
                curr = get_multi_id_range_skip(words,m,i,j,n)
                if curr in skip_dict:
                    if m > 0:
                        m -= 1
                    else:
                        if last_m != -1:
                            if last_m< m:
                                spans.append((last_m,n - 1,i,j))
                        last_m = 0
                        m = i - 1
                        n+= 1
                else:
                    if last_m != -1:
                        if last_m<= m:
                            spans.append((last_m,n - 1,i,j))
                            last_m = m - 1

                    if m == i -1:
                        stop = True
                    else:
                        last_m = m +1
                        m = i - 1
                        n += 1

            if not stop and last_m != -1:
                spans.append((last_m,n - 1,i,j))
            j += 1
    return spans

def get_next_breaks(so_far,span_satisfy_dict,start_index,max_index):
    i = start_index
    while i <= max_index:
        if span_satisfy_dict[i] and not span_satisfy_dict[i].intersection(so_far):
            return span_satisfy_dict[i]
        i += 1
    return set()

def has_subset(curr_set,final_set):
    for compare_set in final_set:
        if curr_set.issuperset(compare_set):
            return True
    return False

def get_all_minimum_breaks_local(so_far, start_index, span_satisfy_dict,max_index):

    todo_list =  [so_far]
    final_set = []
    while todo_list:
        curr = todo_list.pop(0)
        new_breaks = get_next_breaks(curr,span_satisfy_dict,start_index,max_index)
        if not new_breaks:
            if not has_subset(curr,final_set):
                final_set.append(curr)
        else:
            for new_break in new_breaks:
                new = copy.copy(curr)
                new.add(new_break)
                todo_list.append(new)
    return final_set

def get_max_index(spans,curr_index):
    far_left = spans[curr_index +1][1] -1
    new = curr_index + 2
    while new < len(spans) and spans[new][0] <= far_left:
        new += 1
    return new - 2

def compatible_with_skip(span, outer_span,inner_span):
    if len(span) == 2:
        return span_in_span(span,(outer_span[0],inner_span[0])) or span_in_span(span,(inner_span[1],outer_span[1]))
    else:
        return span[0] >= outer_span[0] and span[1] <= outer_span[1] and span[2] <= inner_span[0] and span[3] >= inner_span[1]


def span_internal_break(span,point):
    return span[1] > point >= span[0] +1

def break_span(span,point):
    return (span[0],point), (point,span[1])

def span_in_span(span1,span2):
    return span1[0] >= span2[0] and span1[1] <= span2[1]


def evaluate_breaks_simple_fixed_range(breaks,predictor_dict,start,stop):
    
    total = 0
    for i in range(start,stop):
        j = len(predictor_dict[i]) - 1
        while len(predictor_dict[i][j][-1]) == 4 or not compatible_with_breaks(predictor_dict[i][j][-1], breaks):
            j -= 1
        total += predictor_dict[i][j][0]
    return total

def ngram_only_find_breaks_and_evaluate(ngram_spans,predictor_dict,start,end):
    break_span = (start,end)
    span_satisfy_dict = {}
    given_breaks = set(break_span)
    to_check = set(range(start + 1,end))
    for i in range(len(ngram_spans)):
        if span_in_span(break_span,ngram_spans[i]):
            return evaluate_breaks_simple_fixed_range(break_span,predictor_dict,start,end), set()
        if i < len(ngram_spans) -1 and overlaps(ngram_spans[i],break_span) and overlaps(ngram_spans[i+1],break_span):
            span_satisfy_dict[i] = set(range(ngram_spans[i+1][0],ngram_spans[i][1] + 1))
        for j in range(ngram_spans[i][0]+1,ngram_spans[i][1]):
            to_check.discard(j)
    given_breaks.update(to_check)
    if not span_satisfy_dict:
        all_possible_breaks = [given_breaks]
    else:        
        all_possible_breaks = get_all_minimum_breaks_local(given_breaks, min(span_satisfy_dict), span_satisfy_dict,max(span_satisfy_dict))
    best = None
    best_breaks = None
    for new_breaks in all_possible_breaks:
        score = evaluate_breaks_simple_fixed_range(new_breaks,predictor_dict,start,end)                                           
        if score > best:
           best = score
           best_breaks = new_breaks
    best_breaks = set(range(start + 1,end)).intersection(best_breaks)
    return best, best_breaks
    
    

def evaluate_skip_span(skip_span,ngram_spans,predictor_dict,break_span,span_result_dict):
    total = 0
    if not (break_span[0] < skip_span[2] and break_span[1] > skip_span[3]):
        return None,None
    inner_span = (skip_span[2],skip_span[3])
    if inner_span in span_result_dict:
        total += span_result_dict[inner_span][0]
        breaks = span_result_dict[inner_span][1]
        if breaks == None:
            breaks = set()
        else:
            breaks = copy.copy(breaks)
        breaks.update(inner_span)
    else:
        best_score, breaks = ngram_only_find_breaks_and_evaluate(ngram_spans,predictor_dict,skip_span[2],skip_span[3])
        span_result_dict[inner_span] = [best_score,breaks,False]
        breaks = copy.copy(breaks)
        breaks.update(inner_span)
        total += best_score
    for i in range(break_span[0],skip_span[2]) + range(skip_span[3], break_span[1]):
        j = len(predictor_dict[i]) - 1
        try:
            while not compatible_with_skip(predictor_dict[i][j][-1], break_span,inner_span):
                j -= 1
            total += predictor_dict[i][j][0]
        except:
            total += -9999
            pass
        
    
    return total,breaks       
    
            
    

def evaluate_breaks(breaks,ngram_spans,skip_spans,predictor_dict,max_index,span_result_dict):
    
    breaks = list(breaks)
    breaks.sort()
    total = 0
    all_extra_breaks = {}
    for i in range(len(breaks) - 1):
        break_span = (breaks[i],breaks[i+1])
        if break_span in span_result_dict:
            part_total, extra_breaks,used_skip = span_result_dict[break_span]
            total += part_total
            if extra_breaks:
                all_extra_breaks[break_span] = [extra_breaks,used_skip]     
        else:
            found = False
            part_total = 0
            for span in ngram_spans:
                if span_in_span(break_span,span):
                    for k in range(breaks[i],breaks[i+1]):
                        j = len(predictor_dict[k]) - 1
                        while not compatible_with_breaks(predictor_dict[k][j][-1], break_span):
                            j -= 1
                        part_total += predictor_dict[k][j][0]
                    total += part_total
                    span_result_dict[break_span] = part_total,None,False
                    found = True
                    break
            if found:
                continue

            part_total, extra_breaks = ngram_only_find_breaks_and_evaluate(ngram_spans,predictor_dict,breaks[i],breaks[i+1])
       
            used_skip = False
            for skip_span in skip_spans:
                if span_in_span(break_span,skip_span):
                    temp_result, temp_extra_breaks = evaluate_skip_span(skip_span,ngram_spans,predictor_dict,break_span,span_result_dict)
                    if temp_result > part_total:
                        part_total = temp_result
                        extra_breaks = temp_extra_breaks
                        used_skip = True
            total += part_total
            span_result_dict[break_span] = [part_total,extra_breaks,used_skip]
            all_extra_breaks[break_span] = [extra_breaks,used_skip]              
    return total, all_extra_breaks                   
                    

def combine_ngram_skip(ngram_span_set,skip_spans):
    '''
    The purpose of this function is make the two sets of ngram spans (sequential
    and gap) consistent with each other, with respect to the groupings
    of independent spans within each list (they must line up, so that that the
    segmentation step can deal with them. The returned ngram_span_set and
    skip span_set are equal in length, and (when they exist) the relevant
    spans in each grouping are to some degree overlapping
    '''
    skip_span_set = []
    for i in range(len(ngram_span_set)):
        skip_span_set.append([])
    updated_min_max = {}
    for skip_span in skip_spans:
        stop = False
        i = 0
        while not stop:
            if i == len(ngram_span_set):
                ngram_span_set.append([])
                skip_span_set.append([skip_span])
                updated_min_max[i] = skip_span[0],skip_span[1]
            if i in updated_min_max:
                span_min,span_max = updated_min_max[i]
            else:
                span_min = ngram_span_set[i][0][0]
                span_max = ngram_span_set[i][-1][1]
            if span_min >= skip_span[1]: # insert new span between i and i -1:
                ngram_span_set = ngram_span_set[:i] + [[]] + ngram_span_set[i:]
                skip_span_set = skip_span_set[:i] + [[]] + skip_span_set[i:]
                updated_mix_max_keys = updated_min_max.keys()
                updated_mix_max_keys.sort(reverse=True)
                for key in updated_mix_max_keys:
                    updated_min_max[key+1] = updated_min_max[key]
                updated_min_max[i] = [skip_span[0],skip_span[1]]
                skip_span_set[i].append(skip_span)
                stop = True
            elif span_max > skip_span[0] >= span_min or span_max >= skip_span[1] > span_min or (skip_span[0] < span_min and skip_span[1] > span_max):
                if i+1 in updated_min_max:
                    next_span_min = updated_min_max[i+1][0]
                else:
                    if len(ngram_span_set) > i + 1:
                        next_span_min = ngram_span_set[i+1][0][0]
                    else:
                        next_span_min = skip_span[1]

                if skip_span[1] > next_span_min: # combine spans
                    new_combo_set = ngram_span_set[i] + ngram_span_set[i+1]
                    new_combo_skip_set = skip_span_set[i] + skip_span_set[i+1]
                    new_combo_skip_set.append(skip_span)
                    j = i + 2
                    if j in updated_min_max:
                        next_span_min = updated_min_max[j][0]
                    elif j < len(ngram_span_set):
                        next_span_min = ngram_span_set[j][0][0]
                    while j < len(ngram_span_set) and skip_span[1] > next_span_min:
                        new_combo_set += ngram_span_set[j]
                        new_combo_skip_set += skip_span_set[j]
                        j += 1
                        if j in updated_min_max:
                            next_span_min = updated_min_max[j][0]
                        elif j < len(ngram_span_set):
                            next_span_min = ngram_span_set[j][0][0]
                    
                    ngram_span_set = ngram_span_set[:i] + [new_combo_set] + ngram_span_set[j:]
                    skip_span_set = skip_span_set[:i] + [new_combo_skip_set] + skip_span_set[j:]
                    updated_mix_max_keys = updated_min_max.keys()
                    updated_mix_max_keys.sort(reverse=True)
                    for key in updated_mix_max_keys:
                        if j> key >= i:
                            del updated_min_max[key]
                        elif key >= j:
                            updated_min_max[key-(j - i - 1)] = updated_min_max[key]
                    updated_min_max[i] = [min(new_combo_set[0][0],new_combo_skip_set[0][0]),max(new_combo_set[-1][1],new_combo_skip_set[-1][1])]
                    stop = True
                else:
                    skip_span_set[i].append(skip_span)
                    if skip_span[0] < span_min or skip_span[1] > span_max:
                        updated_min_max[i] = [min(skip_span[0],span_min),max(skip_span[1],span_max)]
                    stop = True
            i+= 1
                
                    
    return ngram_span_set,skip_span_set


def chunk_sentence_by_ngram_decomposition(words,org_pos,cond_prob_dict,skip_cond_prob_dict):
    if len(words) == 0:
        return [],{}

    ngram_spans = get_ngram_spans(words,cond_prob_dict)
    skip_spans = get_skip_spans(words,org_pos,skip_cond_prob_dict)
    ngram_spans, skip_spans = combine_ngram_skip(ngram_spans,skip_spans)
    final_chunks = []
    predictor_dict = {}
    extra_breaks_dict = {}
    for k in range(len(ngram_spans)):
        ngram_span_set = ngram_spans[k]
        skip_span_set = skip_spans[k]
        combined_span_set = ngram_span_set + skip_span_set
        combined_span_set.sort()
        new_combined_span_set = [combined_span_set[0][:2]]
        for i in range(1,len(combined_span_set)):
            if combined_span_set[i][0] == new_combined_span_set[-1][0]:
                new_combined_span_set[-1] = combined_span_set[i][:2]
            elif new_combined_span_set[-1][1] <  combined_span_set[i][1]:
                new_combined_span_set.append(combined_span_set[i][:2])
        combined_span_set = new_combined_span_set            
        span_satisfy_dict = {}
        for i in range(len(combined_span_set)-1):
            span_satisfy_dict[i] = set(range(combined_span_set[i+1][0],combined_span_set[i][1] + 1))
        
        for span in combined_span_set:
            for i in range(span[0],span[1]):
                if i not in predictor_dict:
                    predictor_dict[i] = get_ranking_of_predictors(words,i, ngram_span_set,cond_prob_dict,skip_span_set,skip_cond_prob_dict)
        sub_result_dict = {}
        new_breaks = set()
        if len(combined_span_set) == 1 and combined_span_set[0] not in ngram_span_set:
            score,temp_extra_breaks = evaluate_breaks(set(combined_span_set[0]),ngram_span_set,skip_span_set,predictor_dict,0,sub_result_dict)
            extra_breaks_dict[combined_span_set[0]] = temp_extra_breaks[combined_span_set[0]]

        for i in range(len(combined_span_set)-1):
            if new_breaks:
                last_break = max(new_breaks)
            else:
                last_break = combined_span_set[0][0]
            if last_break in span_satisfy_dict[i]:
                continue

            max_index = get_max_index(combined_span_set,i)
            all_possible_breaks = get_all_minimum_breaks_local(set([last_break]), i, span_satisfy_dict,max_index)
            end_of_relevant_span = combined_span_set[max_index + 1][1]
            best = None
            best_breaks = False
            extra_breaks = None
            for breaks in all_possible_breaks:
                breaks.add(end_of_relevant_span)
                score,temp_extra_breaks = evaluate_breaks(breaks,ngram_span_set,skip_span_set,predictor_dict,max_index,sub_result_dict)
                if score > best:
                    best = score
                    best_breaks = breaks
                    best_extra = temp_extra_breaks
            best_breaks.remove(last_break)
            new_break = min(best_breaks)
            new_breaks.add(new_break)
            new_span =  (last_break, new_break) 
            if best_extra and new_span in best_extra:
                extra_breaks_dict[new_span] = best_extra[new_span]
            if new_break in span_satisfy_dict[len(combined_span_set)-2]:
                final_span = (new_break,end_of_relevant_span)
                if final_span in best_extra:
                    extra_breaks_dict[final_span] = best_extra[final_span]
        new_spans = []        
        span_set = combined_span_set
        for new_break in new_breaks:
            for span in span_set:
                if span_internal_break(span,new_break):
                    found_dub1 = False
                    found_dub2 = False
                    span1,span2 = break_span(span,new_break)
                    for i in range(len(span_set)):
                        if span_set[i] != span:
                            if span_in_span(span1, span_set[i]):
                                found_dub1 = True
                            if span_in_span(span2, span_set[i]):
                                found_dub2 = True
                    if not found_dub1 and span1[1] - span1[0] > 1:
                        new_spans.append(span1)
                    if not found_dub2 and span2[1] - span2[0] > 1:
                        new_spans.append(span2)
                else:
                    new_spans.append(span)
            span_set = new_spans
            new_spans = []
        final_chunks.extend(span_set)
    final_chunks.sort()
    return final_chunks,extra_breaks_dict



def segment_corpus_by_decomposition(start,stop,cond_prob_dict,skip_cond_prob_dict,rev_id_dict,POS_rev_id_dict,Q):
    for sentence in read_sentence_from_corpus(options.corpus,start,stop):
        new_vocab = []
        new_skip_vocab = []
        pos = [pair[1] for pair in sentence]
        words = [pair[0] for pair in sentence]
        words = replace_words_with_id(words,rev_id_dict)
        org_pos = copy.copy(pos)
        pos = replace_words_with_id(pos,POS_rev_id_dict)
        chunked_spans,extra_breaks = chunk_sentence_by_ngram_decomposition(words,org_pos, cond_prob_dict,skip_cond_prob_dict)
        taken_indicies = set()
        for span in chunked_spans:
            if span in extra_breaks:               
                breaks = list(extra_breaks[span][0])
                breaks.sort()
                if extra_breaks[span][1]:
                    new_skip_vocab.append(get_multi_id_range_skip(words,span[0],breaks[0],breaks[-1],span[1]))
                    for i in range(len(breaks) - 1):
                        index = get_multi_id_range(words,breaks[i],breaks[i+1])
                        if index != 0:
                            new_vocab.append(index)
                else:
                    index = get_multi_id_range(words,span[0],breaks[0])
                    if index != 0:
                        new_vocab.append(index)
                    index = get_multi_id_range(words,breaks[-1], span[0])
                    if index != 0:
                        new_vocab.append(index)
                    for i in range(len(breaks) - 1):
                        index = get_multi_id_range(words,breaks[i],breaks[i+1])
                        if index != 0:
                            new_vocab.append(index)                                                      
            else:
                index = get_multi_id_range(words,span[0],span[1])
                new_vocab.append(get_multi_id_range(words,span[0],span[1]))
            taken_indicies.update(range(span[0],span[1]))
        for i in range(len(words)):
            if i not in taken_indicies and words[i] != -2:
                new_vocab.append(words[i])
        for vocab in new_vocab:
            Q.put(vocab)
        for vocab in new_skip_vocab:
            Q.put(-vocab)
    Q.put(-1)

f = open("temp_options.dat","rb")
options = cPickle.load(f)
f.close()
if options.lang == "en":
    matches_gap = matches_gap_en
else:
    matches_gap = matches_gap_fr

if __name__ == "__main__":
    f = open("temp_ngrams.dat","rb")
    id_dict = cPickle.load(f)
    cPickle.load(f)
    cPickle.load(f)
    token_count = cPickle.load(f)
    sentence_count = cPickle.load(f)
    POS_id_dict = cPickle.load(f)
    f.close()

    f = open("temp_LPR_stats.dat", "rb")
    cond_prob_dict =  cPickle.load(f)
    skip_cond_prob_dict = cPickle.load(f)
    f.close()

    rev_id_dict = {}
    for ID in id_dict:
        rev_id_dict[id_dict[ID]] = ID

    POS_rev_id_dict = {}
    for ID in POS_id_dict:
        POS_rev_id_dict[POS_id_dict[ID]] = ID
    
    processes = []
    Q = Queue()

    intervals = []
    for i in range(options.workers):
        intervals.append(i*(sentence_count/options.workers))
    intervals.append(sentence_count)

    for i in range(options.workers):
        processes.append(Process(target=segment_corpus_by_decomposition,args=(intervals[i],intervals[i+1],cond_prob_dict,skip_cond_prob_dict,rev_id_dict,POS_rev_id_dict,Q)))
        processes[-1].start()


    initial_vocab = {}
    initial_skip_vocab = {}

    done_threads = 0
    while done_threads < options.workers:
        result = Q.get()
        if result == -1:
            done_threads += 1
        else:
            if result> 0:
                initial_vocab[result] = initial_vocab.get(result,0) + 1
            else:
                initial_skip_vocab[-result] = initial_skip_vocab.get(-result,0) + 1

    fout = open("temp_initial_vocab.dat","wb")
    cPickle.dump(initial_vocab,fout,-1)
    cPickle.dump(initial_skip_vocab,fout,-1)
    fout.close()


